/*
  Program:           BROWSE.C
    
  Author:            James L. Dean
                     406 40th Street
                     New Orleans, LA 70124
      
  Date Written:      September 10, 1984
       
  Language:          Microsoft C 4.0
       
  Operating System:  MS-DOS 2.11

  Machine:           Zenith Data Systems Corporation Z-158

  Microprocessor:    Intel Corporation 8088

  Remarks.

      This program uses a fractionating cipher to display the results of
  deciphering a file without changing the file.

      For each sector (128 bytes) of the file, sixteen substitutions of
  bytes for bytes in the sector are alternated with sixteen permutations of
  the half bytes in the sector.

      Only one permutation pattern and one substitution pattern are used.
  These patterns are fixed by a thirty (or fewer) character key.
*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <dos.h>

#define TRUE 1
#define FALSE 0

static void clear_patterns(int *,int *);
static void clear_random_number_generator(int *,int *);
static void decipher_each_sector(int *,int *);
static void display_half_bytes(int *,int *);
static void generate_patterns(int *,int *);
static void generate_permutation_pattern(int *,int *,int **,int **,int *);
static void generate_substitution_pattern(int *,int *,int **,int **,int *);
static void get_key(char *);
static void permute_half_bytes(int *,int *,int *);
static int  random_number(int *,int *,int **,int **);
static void read_key(char *);
static void seed_random_number_generator(char *,int *,int *,int **,int **);
static void set_up_half_byte_translation(int *,int *);
static void substitute_bytes(int *,int *,int *,int *,int *);
static void translate_sector_to_half_bytes(unsigned char *,int *,int *,int *);

FILE *disk_file;

void main(argc,argv)
 int  argc;
 char *argv[];
  {
    static int permutation [256];
    static int substitution [256];

    if (argc < 2)
      printf("Usage:  BROWSE <file descriptor>\n");
    else
      if ((disk_file=fopen(argv[1],"rb")) == NULL)
        printf("? Cannot open %s for input\n",argv[1]);
      else
        {
          generate_patterns(&permutation[0],&substitution[0]);
          decipher_each_sector(&permutation[0],&substitution[0]);
          fclose(disk_file);
          clear_patterns(&permutation[0],&substitution[0]);
        }
    return;
  }

static void read_key(key)
 char *key;
  {
    static union REGS inreg;
    register int      key_length;
    static union REGS outreg;
    register int      tem_char;

    key_length=0;
    key[0]='\0'; 
    do
      {
        inreg.h.ah=(short) 8;
        intdos(&inreg,&outreg);
        tem_char=(int) outreg.h.al;
        if (tem_char != 13)
          {
            if (tem_char == 8)
              {
                if (key_length > 0)
                  key[--key_length]='\0';
              }
            else
              if (key_length < 255)
                {
                  key[key_length++]=(char) tem_char;
                  key[key_length]='\0';
                }
              else 
                putchar(7);
          }
      }
    while (tem_char != 13);
  }

static void get_key(key_1)
  char *key_1;
    {
      register int key_index;
    /*
          Obtain the thirty character key.  Lowercase characters are
      converted to uppercase.
    */
      printf("\n");       
      printf("Key?  ");
      read_key(key_1);
      while (strlen(key_1) < 30) strcat(key_1," ");
      printf("\n");
      for (key_index=0; key_index < 30; key_index++)
        key_1[key_index]=toupper(key_1[key_index]);
      return;
    }

static void seed_random_number_generator(key_1,previous_random_number,
 r_n_partial_sum,add_ptr,replace_ptr)
  char *key_1;
  int  *previous_random_number;
  int  *r_n_partial_sum;
  int  **add_ptr;
  int  **replace_ptr;
    {
      register int key_index;
      static   int *previous_r_n_ptr;
    /*
           Use the position in the collating sequence of each character
      of the key phrase to supply seeds for the pseudo-random number
      generator.
    */
      previous_r_n_ptr=previous_random_number;
      while (*key_1)
        {
          *previous_r_n_ptr=(int) *key_1;
          previous_r_n_ptr++;
          *key_1='\000';
          key_1++;
        }
      *replace_ptr=previous_random_number;
      *r_n_partial_sum=0;
      for (key_index=0; key_index < 29; key_index++)
        {
          *r_n_partial_sum+=*previous_random_number;
          if (*r_n_partial_sum >= 257)
            *r_n_partial_sum-=257;
          previous_random_number++;
        }
      *add_ptr=previous_random_number;
      return;
    }

static int random_number(previous_random_number,r_n_partial_sum,add_ptr,
 replace_ptr)
  int *previous_random_number;
  int *r_n_partial_sum;
  int **add_ptr;
  int **replace_ptr;
   {
     register int r_n;
    /*
        Each pseudo-random number is the modulo sum of the
      previous thirty pseudo-random numbers.  A prime modulus
      makes it likely that the pseudo-random numbers will be
      uniformly distributed.  To speed computation, a partial
      sum of 29 of the 30 previous pseudo-random numbers is
      maintained.
    */
      do
        {
          r_n=*r_n_partial_sum+(*(*add_ptr));
          if (r_n >= 257) r_n-=257;
          *r_n_partial_sum=r_n-(*(*replace_ptr));
          if (*r_n_partial_sum < 0) *r_n_partial_sum+=257;
          *(*replace_ptr)=r_n;
          *add_ptr=*replace_ptr;
          (*replace_ptr)++;
          if (*replace_ptr == (previous_random_number+30))
            *replace_ptr=previous_random_number;
        }
      while (r_n > 255);
      return(r_n);
    }

static void generate_substitution_pattern(previous_random_number,
 r_n_partial_sum,add_ptr,replace_ptr,substitution)
  int *previous_random_number;
  int *r_n_partial_sum;
  int **add_ptr;
  int **replace_ptr;
  int *substitution;
    {
      static   int byte_value;
      static   int encipherment [256];
      register int index_1;
      register int index_2;
      static   int tem;

      for (byte_value=0; byte_value < 256; byte_value++)
        encipherment[byte_value]=byte_value;
      for (index_1=0; index_1 < 256; index_1++)
        {
          index_2=random_number(previous_random_number,
           r_n_partial_sum,add_ptr,replace_ptr);
          tem=encipherment[index_1];
          encipherment[index_1]=encipherment[index_2];
          encipherment[index_2]=tem;
        }
      tem=0;
      for (index_1=0; index_1 < 256; index_1++)
        *(substitution+encipherment[index_1])=index_1;
      return;
    }

static void generate_permutation_pattern(previous_random_number,
 r_n_partial_sum,add_ptr,replace_ptr,permutation)
  int *previous_random_number;
  int *r_n_partial_sum;
  int **add_ptr;
  int **replace_ptr;
  int *permutation;
    {
      static   int byte_value;
      register int index_1;
      register int index_2;
      static   int *ptr_1;
      static   int *ptr_2;
      static   int tem;

      for (byte_value=0; byte_value < 256; byte_value++)
        *(permutation+byte_value)=byte_value;
      ptr_1=permutation;
      for (index_1=0; index_1 < 256; index_1++)
        {
          index_2=random_number(previous_random_number,
           r_n_partial_sum,add_ptr,replace_ptr);
          ptr_2=permutation+index_2;
          tem=*ptr_1;
          *ptr_1=*ptr_2;
          *ptr_2=tem;
          ptr_1++;
        }
      tem=0;
      return;
    }

static void clear_random_number_generator(previous_random_number,
 r_n_partial_sum)
  int *previous_random_number;
  int *r_n_partial_sum;
    {
      register int add_index;

      for (add_index=0; add_index < 30; add_index++)
        {
          *previous_random_number=0;
          previous_random_number++;
        }
      *r_n_partial_sum=0;
      return;
    }

static void generate_patterns(permutation,substitution)
  int *permutation;
  int *substitution;
    {
      static   int  *add_ptr;
      static   char key_1 [256];
      static   int  previous_random_number [30];
      static   int  r_n_partial_sum;
      static   int  *replace_ptr;

      get_key(key_1);
      seed_random_number_generator(key_1,&previous_random_number[0],
       &r_n_partial_sum,&add_ptr,&replace_ptr);
      generate_substitution_pattern(&previous_random_number[0],
       &r_n_partial_sum,&add_ptr,&replace_ptr,substitution);
      generate_permutation_pattern(&previous_random_number[0],
       &r_n_partial_sum,&add_ptr,&replace_ptr,permutation);
      clear_random_number_generator(&previous_random_number[0],
       &r_n_partial_sum);
      return;
    }

static void set_up_half_byte_translation(high_order_half_byte,
 low_order_half_byte)
  int *high_order_half_byte;
  int *low_order_half_byte;
    {
      register int half_byte_index_1;
      register int half_byte_index_2;

      for (half_byte_index_1=0; half_byte_index_1 < 16; half_byte_index_1++)
        for (half_byte_index_2=0; half_byte_index_2 < 16; half_byte_index_2++)
          {
            *low_order_half_byte=half_byte_index_2;
            *high_order_half_byte=half_byte_index_1;
            low_order_half_byte++;
            high_order_half_byte++;
          }
      return;
    }

static void translate_sector_to_half_bytes(sector,text_1_half_byte,
 low_order_half_byte,high_order_half_byte)
  unsigned char *sector;
  int           *text_1_half_byte;
  int           *low_order_half_byte;
  int           *high_order_half_byte;
    {
      register int byte_in_sector;
      register int byte_value;

      for (byte_in_sector=0; byte_in_sector < 128; byte_in_sector++)
        {
          byte_value=(int) *sector;
          *text_1_half_byte=*(high_order_half_byte+byte_value);
          text_1_half_byte++;
          *text_1_half_byte=*(low_order_half_byte+byte_value);
          text_1_half_byte++;
          sector++;
        }
      return;
    }

static void permute_half_bytes(permutation,text_2_half_byte,text_1_half_byte)
  int *permutation;
  int *text_2_half_byte;
  int *text_1_half_byte;
    {
      register int half_byte_index;

      for (half_byte_index=0; half_byte_index < 256; half_byte_index++)
        {
          *text_1_half_byte=*(text_2_half_byte+(*permutation));
          permutation++;
          text_1_half_byte++;
        }
      return;
    }

static void substitute_bytes(substitution,text_1_half_byte,text_2_half_byte,
 low_order_half_byte,high_order_half_byte)
  int *substitution;
  int *text_1_half_byte;
  int *text_2_half_byte;
  int *low_order_half_byte;
  int *high_order_half_byte;
    {
      static   int byte_in_sector;
      register int byte_value;
      register int offset;

      for (byte_in_sector=0; byte_in_sector < 128; byte_in_sector++)
        {
          offset=*text_1_half_byte;
          offset*=16;
          text_1_half_byte++;
          offset+=*text_1_half_byte;
          text_1_half_byte++;
          byte_value=*(substitution+offset);
          *text_2_half_byte=*(high_order_half_byte+byte_value);
          text_2_half_byte++;
          *text_2_half_byte=*(low_order_half_byte+byte_value);
          text_2_half_byte++;
        }
      return;
    }

static void display_half_bytes(text_1_half_byte,end_of_file)
  int *text_1_half_byte;
  int *end_of_file;
    {
      register int byte_in_sector;
      register int byte_value;

      byte_in_sector=0;
      while ((byte_in_sector < 128) && (! *end_of_file))
        {
          byte_value=*text_1_half_byte;
          text_1_half_byte++;
          byte_value*=16;
          byte_value+=*text_1_half_byte;
          text_1_half_byte++;
          if (byte_value == 26)
            *end_of_file=TRUE;
          else
            {
              putchar(byte_value);
              byte_in_sector++;
            }
        }
      return;
    }

static void decipher_each_sector(permutation,substitution)
  int *permutation;
  int *substitution;
    {
      static   int           byte_index;
      static   int           end_of_file;
      static   int           half_byte_index;
      static   int           high_order_half_byte [256];
      register int           iteration;
      static   int           low_order_half_byte [256];
      register int           num_bytes;
      static   unsigned char sector [128];
      static   int           text_1_half_byte [256];
      static   int           text_2_half_byte [256];

      set_up_half_byte_translation(
       &high_order_half_byte[0],&low_order_half_byte[0]);
      end_of_file=FALSE;
      printf("\n");
      while ((! end_of_file) && ((num_bytes=
       fread((char *) &sector[0],1,128,disk_file)) > 0))
        {
          translate_sector_to_half_bytes(&sector[0],&text_1_half_byte[0],
           &low_order_half_byte[0],&high_order_half_byte[0]);
        /*
                For a faster though less secure cipher, replace the next
          line with "for (iteration=1; iteration <=2; iteration++)".
        */
          for (iteration=1; iteration <= 16; iteration++)
            {
              substitute_bytes(
               substitution,&text_1_half_byte[0],&text_2_half_byte[0],
               &low_order_half_byte[0],&high_order_half_byte[0]);
              permute_half_bytes(
               permutation,&text_2_half_byte[0],&text_1_half_byte[0]);
            } 
          display_half_bytes(&text_1_half_byte[0],&end_of_file);
        }
      for (byte_index=0; byte_index < 128; byte_index++)
        sector[byte_index]=0;
      for (half_byte_index=0; half_byte_index < 256; half_byte_index++)
        {
          text_1_half_byte[half_byte_index]=0;
          text_2_half_byte[half_byte_index]=0;
        }
      return;
    }

static void clear_patterns(permutation,substitution)
  int *permutation;
  int *substitution;
    {
      static int index_1;

      for (index_1=0; index_1 < 256; index_1++)
        {
          *permutation=0;
          *substitution=0;
          permutation++;
          substitution++;
        }
      return;      
    }
